import numpy as np


# wrap list as numpy array
def asarray(func):
    """
        Decorator to wrap a function that returns
        a list to a numpy array.
    """
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return np.array(result)
    return wrapper

# sort a list - something wrong with this
def sortresult(func):
    """
        Decorator to wrap a function that returns
        a list to return a sorted list.
    """
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return sorted(result)
    return wrapper

def constant(func):
    """
        Decorator that prevents a value from being changed.
        Represents a constant as best as possible.
    """
    def fset(self, value):
        raise TypeError("Constant - no setting!")

    def fget(self):
        return func(self)
    return property(fget, fset)



"""
    The module that includes all core calculation

    Includes binning lines into histograms.
"""
import math
import numpy as np
from typing import Dict, List, Tuple

from .exceptions import (
    UnphysicalValueException,
    UnknownOrUnstableNuclideException,
    NoDataException,
)
from .database import ReadOnlyDatabase
from .decorators import asarray
from .inventory import UnstablesInventory


LOG_TWO_BASE_E = math.log(2)

linspace = np.linspace
logspace = np.logspace


class EnergyGrid:
    """
    This represents a simple energy grid to define
    the bins for the gamma spec.

    It is basically a numpy array with some units
    and some protection on negative values.
    """

    __slots__ = ["bounds"]

    def __init__(self, bounds: np.ndarray = linspace(0.0, 10e6, 10000)):
        """
        Energies in eV

        TODO: use a library to handle units

        :param bounds: a numpy array defining the binning
        """

        if np.any(bounds < 0):
            raise UnphysicalValueException("Energies cannot be negative.")

        self.bounds = bounds

    def __len__(self) -> int:
        """
        The number of energy bounds

        :return: the number of energy bounds in the grid
        """
        return len(self.bounds)

    def __getitem__(self, i: int) -> float:
        """
        Get the i'th item in the data (energy bounds)

        Does no bound checking on index

        :param i: the index in the grid to access
        :returns: the value at the given index
        """
        return self.bounds[i]

    def __str__(self) -> str:
        """
        A string representation of the energy bounds

        :returns: a string representing the energy bounds
        """
        return str(self.bounds)

    @property
    def nrofbins(self) -> int:
        """
        The number of energy bins, equal to the
        number of energy bounds - 1

        :returns: the number of bins
        """
        return len(self) - 1

    @property
    def units(self) -> str:
        """
        Return the unit type as a string
        TODO: support other units

        :returns: a string representing the units of energy
        """
        return "eV"

    @property
    @asarray
    def midpoints(self) -> np.array:
        """
        Return a numpy array of the midpoint values, in eV

        :returns: a numpy array of the midpoint values in eV
        """
        return (self.bounds[:-1] + self.bounds[1:]) / 2

    @property
    def minEnergy(self) -> float:
        """
        Return the minimum energy, in eV, of the energy grid

        :returns: the minimum energy, in eV, of the energy grid
        """
        return np.min(self.bounds)

    @property
    def maxEnergy(self) -> float:
        """
        Return the maximum energy, in eV, of the energy grid

        :returns: the maximum energy, in eV, of the energy grid
        """
        return np.max(self.bounds)


class LineAggregator:
    """
    A simple class for reading lines of a single decay type
    i.e. "gamma" and binning them in appropriate bins
    according to the energy grid definition.
    """

    __slots__ = ["db", "grid", "lines", "values"]

    def __init__(self, db: ReadOnlyDatabase, grid: EnergyGrid):
        self.db = db

        self.grid = grid

        # here we just store lines and intensities
        self.lines = []

        # NOTE: values are intensities multiplied by the activity of each
        # nuclide here
        self.values = []

    def _sortlines(self):
        # sort the lines in ascending energy to make it easier to bin in a
        # histogram
        sorteddata = sorted(zip(self.lines, self.values), key=lambda pair: pair[0])
        self.lines = [x for x, _ in sorteddata]
        self.values = [y for _, y in sorteddata]

    def _findlines(
        self, inventory: UnstablesInventory, *args, spectype: str = "gamma", **kwargs
    ):
        lines = []
        values = []

        for zai, activity in inventory:
            name = self.db.getname(zai)
            # check it exists in database
            if name not in self.db:
                raise UnknownOrUnstableNuclideException(
                    "{} not in database - maybe too exotic or is it stable?".format(zai)
                )

            # check that data exists for that decay type
            if spectype not in self.db.gettypes(name):
                raise NoDataException(
                    "{} does not have {} decay mode".format(name, spectype)
                )

            lines.extend(self.db.getenergies(name, spectype=spectype))
            values.extend(self.db.getintensities(name, spectype=spectype) * activity)

        return lines, values

    def _makehist(self, *args, **kwargs):
        hist = np.zeros(self.grid.nrofbins)

        if len(self.lines) > 0:
            # sort the lines in ascending energy to make it easier to bin in a
            # histogram
            self._sortlines()

            # loop over lines to find appropriate bin
            ibin = 0
            for i, line in enumerate(self.lines):
                # loop over bounds from the last found bin
                for j in range(ibin, len(self.grid) - 1):
                    if line >= self.grid[j] and line < self.grid[j + 1]:
                        hist[j] += self.values[i]
                        ibin = j
                        # once we've found the line break the inner loop
                        break

        return hist, self.grid.bounds

    def __call__(
        self, inventory: UnstablesInventory, *args, spectype: str = "gamma", **kwargs
    ):
        """
        Gets the lines from the full inventory

        throws an exception if nuclide is stable or is not in database
        """
        self.lines, self.values = self._findlines(
            inventory, *args, spectype=spectype, **kwargs
        )

        return self._makehist(*args, **kwargs)


class LineAverageEnergyAggregator(LineAggregator):
    """
    Instead of LineAggregator which just bins the intensities
    times the activity, the LineAverageEnergyAggregator first
    scales the result by line_energy/midpoint_bin_energy which
    is what FISPACT-II does and is done so to conserve energy.
    This is important for dose calculations.

    For very fine energy grids (high number of bins) this will
    have little effect and should produce results similar to
    LineAggregator
    """

    def _makehist(self, *args, **kwargs):
        hist = np.zeros(self.grid.nrofbins)

        if len(self.lines) > 0:
            self._sortlines()

            average_energies = self.grid.midpoints

            # loop over lines to find appropriate bin
            ibin = 0
            for i, line in enumerate(self.lines):
                # loop over bounds from the last found bin
                for j in range(ibin, len(self.grid) - 1):
                    if line >= self.grid[j] and line < self.grid[j + 1]:
                        # how to handle zero average energy?
                        # we scale the values by the line energy/average bin energy
                        # in order to conserve energy for dose calculations
                        hist[j] += self.values[i] * line / average_energies[j]
                        ibin = j
                        break

        return hist, self.grid.bounds


class MultiTypeLineAggregator(LineAggregator):
    """
    A simple class for reading lines of multiple decay type
    i.e. ["gamma", "x-ray"] and binning them in appropriate bins
    according to the energy grid definition.

    Same as LineAggregator but supports multiple types
    of spectra - gamma + x-ray +beta for example.
    """

    def __call__(
        self,
        inventory: UnstablesInventory,
        *args,
        types: List[str] = ["gamma", "x-ray"],
        **kwargs
    ):
        """
        Gets the lines from the full inventory

        throws an exception if nuclide is stable or is not in database
        """
        for spectype in types:
            lines, values = self._findlines(
                inventory, *args, spectype=spectype, **kwargs
            )
            self.lines.extend(lines)
            self.values.extend(values)

        return self._makehist(*args, **kwargs)


def get_zai_props(db: ReadOnlyDatabase, nuc: str) -> Tuple[int, int, int]:
    # Z, A, I
    """
    Returns the charge (Z), atomic mass number (A) and the isomeric state (I).

    Nuclide must exist in data and must be unstable.

    :param db: The database to access halflife data used in the conversion
    :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
    No spaces and case sensitive!
    :returns: the (Z, A, I) tuple
    """
    zai = db.getzai(nuc)
    return math.floor(zai / 10000), math.floor(zai / 10) % 1000, zai % 10


def activity_from_atoms(db: ReadOnlyDatabase, nuclide: str, atoms: float) -> float:
    """
    Returns activity (Bq) given a number of atoms (typical in FISPACT-II or alike).

    Nuclide must exist in data and must be unstable.

    :param db: The database to access halflife data used in the conversion
    :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
    No spaces and case sensitive!
    :param atoms: the number of atoms for the given nuclide
    :returns: the activity (Bq) for the given radionuclide
    """
    return LOG_TWO_BASE_E * atoms / db.gethalflife(nuclide)


def atoms_from_activity(db: ReadOnlyDatabase, nuclide: str, activity: float) -> float:
    """
    Returns the number of atoms given an activity (Bq).

    Nuclide must exist in data and must be unstable.

    :param db: The database to access halflife data used in the conversion
    :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
    No spaces and case sensitive!
    :param activity: the activity (Bq) for the given radionuclide
    :returns: the number of atoms for the given radionuclide
    """
    return db.gethalflife(nuclide) * activity / LOG_TWO_BASE_E


def make_inventory_from_atoms(
    db: ReadOnlyDatabase, atoms_inv: Dict[int, float]
) -> UnstablesInventory:
    """
    A factory function to create an inventory from the number
    of atoms instead of activities.
    Each tuple should be a ZAI and atoms.
    i.e. [(10030, 4.5e8), (20040, 2.2321e4), ...]
    """
    return UnstablesInventory(
        [(k, atoms_from_activity(db, db.getname(k), v)) for k, v in atoms_inv.items()]
    )



"""
    A module for trying to identify nuclides
    from binned lines
"""
import numpy as np


from .database import ReadOnlyDatabase, sortedlines
from .core import EnergyGrid


class BinWiseNuclideIdentifier():
    """
        A very simple nuclide identifier.

        It loops over bins (1d-array like) with a
        corresponding energy grid to see what nuclides
        can be found in that range.

        TODO: Support a tolerance to energies in order
        to cater for uncertainty in measurement of energy.

        Parameters
        ----------
        db: The spectral database

        Attributes
        ----------
        grid: The energy grid
        nuclides: A list containing the nuclides identified in 
        the spectrum per bin

    """
    def __init__(self, db: ReadOnlyDatabase):
        self.db = db
        self.nuclides = []

    def __call__(self, values: np.ndarray, grid: EnergyGrid, *args,
                 excludes: list = None, spectype: str = "gamma", 
                 progress: bool = True, **kwargs):
        """
            Finds the nuclides in the histogram.

            Only supports single spectrum type.

            TODO: Support multi type spectrum i.e. "gamma" + "x-ray"

            Parameters
            ----------
            excludes: ignore a list of nuclides, that we know should not 
            be in the spectrum
        """
        self.nuclides = []

        # make sure that data is matching
        assert len(values) == grid.nrofbins

        if excludes is None or type(excludes) != list:
            excludes = []

        iterable = lambda x: x
        if progress:
            try:
                from tqdm import tqdm
                iterable = lambda x: tqdm(x)
            except:
                print("Progress not possible without tqdm.")

        # get all lines of that type from the database and sort them in ascending energy
        sorteddata = sortedlines(self.db, spectype=spectype, byenergy=True)

        lastindex = 0
        # get lower and upper energy bounds
        # loop through hist
        for ihist, (lb, ub) in iterable(enumerate(zip(grid.bounds[:-1], grid.bounds[1:]))):

            # the potential nuclides in the current bin
            nucs = []

            # only if the value is non zero
            if values[ihist] > 0:
                # loop over database lines
                for iline, (nuc, energy) in enumerate(sorteddata[lastindex:]):

                    # if line energy is greater than hist bound - skip to next bin
                    if energy > ub:
                        break

                    lastindex = iline
                    # found a nuclide
                    if (energy >= lb) and (energy < ub) and (nuc not in excludes):
                        nucs.append((nuc, energy))

            self.nuclides.append(nucs)

        return self.nuclides



"""
    A module for defining inventories of unstable nuclides
"""
import collections
from typing import List, Tuple

from .exceptions import UnphysicalValueException


class UnstablesInventory:
    """
    A simple data structure to represent inventory data.
    A list of zais and activities (Bq).

    Note that this should only be used for inventories with
    non zero activities i.e. only unstables, since stables have
    no spectrum.

    NOTE: This does not check if ZAI is unstable (TODO: should we do this?).

    An ordered dict is used to preserve the
    order to which items where added.

    Data should be a list of tuples
    with nuclide ZAI and activity
    i.e.
        [(10030, 4.5e8), (20040, 2.2321e4), ...]

    throws UnphysicalValueException if activities are 0 <=
    """

    __slots__ = ["_raw"]

    def __init__(self, data=None):
        self._raw = collections.OrderedDict()
        if data:
            for z, a in data:
                if a <= 0:
                    raise UnphysicalValueException(
                        "Only supports unstable nuclides, activity must be positive."
                    )
                self._raw[z] = self._raw.get(z, 0) + a

    def append(self, zai: int, activity: float):
        if isinstance(zai, int) and isinstance(activity, float):
            if activity <= 0:
                raise UnphysicalValueException(
                    "Only supports unstable nuclides, activity must be positive."
                )
            self._raw[zai] = self._raw.get(zai, 0) + activity
        else:
            raise TypeError("Expects ZAI as integer and activity as float.")

    def reset(self) -> None:
        self._raw = collections.OrderedDict()

    def __len__(self) -> int:
        return len(self._raw.items())

    def __getitem__(self, index) -> Tuple[int, float]:
        return list(self._raw.items())[index]

    def findactivitybyzai(self, zai) -> float:
        """
        Given a ZAI value find the corresponding activity (Bq).
        If zai is not valid or is not in inventory, return 0.0
        """
        return self._raw.get(zai, 0)

    @property
    def zais(self) -> List[int]:
        return list(self._raw.keys())

    @property
    def activities(self) -> List[float]:
        """
        In Bq
        """
        return list(self._raw.values())



"""
    A set of utility functions for ActiGamma
"""
from pprint import pprint


# function to plot histogram
def getplotvalues(bounds, binvalues):
    """
        Matplotlib hist is slow. Use a hist like plot with conventional line plot.

        Expects N+1 bounds where len(binvalues) = N
        and returns two arrays both of length 2N to plot as line plot.

        Make sure arrays are of correct length
        len(binvalues) + 1 == len(bounds)

        :param bounds: the X values of the histogram (bounds)
        :param binvalues: the Y values of the histogram (bin values)
        :return: matching length arrays as a tuple
    """
    assert (len(binvalues) + 1) == len(bounds)

    plotbounds, plotvalues = [], []
    for i, value in enumerate(binvalues):
        plotbounds.extend([bounds[i], bounds[i + 1]])
        plotvalues.extend([value, value])

    return plotbounds, plotvalues



"""
    All ActiGamma exception definitions
"""


class ActiGammaException(Exception):
    """
        Base exception for all ActiGamma exceptions
    """


class AbstractClassException(ActiGammaException):
    """
        Exception for all intantiating abstract classes
    """


class UnphysicalValueException(ActiGammaException):
    """
        Exception for all using or defining unphysical values
        such as negative mass or energy
    """


class UnknownOrUnstableNuclideException(ActiGammaException):
    """
        Exception for a nuclide that is unknown in the database
    """


class NoDataException(ActiGammaException):
    """
        Exception for missing data in the database
    """



"""
    The database definition and facades.

    Includes a precooked JSON loader for the in built decay 2012 data and
    allows extension for other database types.
"""
import os
import json
from typing import List, Tuple

from .decorators import asarray, constant, sortresult
from .exceptions import AbstractClassException

# hacky but will do, the database is one large JSON file with line data
# we load it into a static data structure which is our database
__RAW_DATABASE_DECAY_2012_FILE__ = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "data", "lines_decay_2012.min.json"
)


class DatabaseJSONFileLoader:
    """
    Context manager to handle JSON datafile loader

    Reads in data as a JSON file into a dictionary

    Expects data in schema as:
    {
        'H3': {
            'halflife': 389105000.0,
            'zai': 10030,
            'beta': {
                'lines': {
                    'energies': [18571.0],
                    'energies_unc': [6.0],
                    'intensities': [1.0],
                    'intensities_unc': [0.0],
                    'norms': [1.0],
                    'norms_unc': [0.0]
                },
                'mean_energy': 5707.4,
                'mean_energy_unc': 1.84397,
                'mean_normalisation': 1.0,
                'mean_normalisation_unc': 0.0,
                'number': 1
            },
        }
    }

    Keys are nuclide name, and each nuclide has a halflife and a zai.
    Following keys are then decay mode type: 'gamma', 'x-ray', ...
    with spectral data
    """

    __slots__ = ["filename"]

    def __init__(self, datafile: str = __RAW_DATABASE_DECAY_2012_FILE__):
        """
        Construct the object given a filename to a raw JSON file

        :param datafile: The filename of the raw JSON file
        containing the spectral data
        """
        self.filename = datafile

    def __enter__(self):
        """
        Opens the file and returns the data as a dictionary
        """
        with open(self.filename, "rt") as fjson:
            return json.loads(fjson.read())

    def __exit__(self, *args):
        """
        Does nothing
        """


ABSTRACT_STR_ERROR = "ReadOnlyDatabase should not be instantiated - please extend with your own database."


class ReadOnlyDatabase:
    """
    A base level read only database for interacting with data

    This assumes the datasource is a JSON file loader, but as long
    as it is a context manager it can be anything.

    ```
    with datasource as source:
        # do something with source here
        ...
    ```
    """

    __slots__ = ["__raw"]

    def __init__(self, datasource=DatabaseJSONFileLoader()):
        """
        Construct the database given a datasource

        ```
        with datasource as source:
            # do something with source here
            ...
        ```

        :param datasource: context manager to load data into database
        """
        self.__raw = {}
        if datasource:
            with datasource as db:
                self.__raw = db

    def __contains__(self, nuclide: str) -> bool:
        """
        Check if nuclide exists in database
        Abstract method - must be extended.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :returns: boolean - true if in database, false otherwise
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    @constant
    def raw(self):
        """
        Get the underlying datastructure - read only.
        Allows users to perform own custom queries, but protected to be read only.
        No setting permitted.

        :returns: a dictionary object representing the underlying data
        """
        return self.__raw

    @property
    def alltypes(self) -> [str]:
        """
        Return all possible unique decay mode types for the whole database
        Abstract method - must be extended.

        :returns: a list of strings representing the list of decay types
        in the database
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    def haslines(self, nuclide: str, spectype: str = "gamma") -> bool:
        """
        Return true if nuclide has spectral information, false otherwise
        Abstract method - must be extended.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: boolean based on if spectral data for type exists
        and has data in database
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    @property
    def allnuclides(self) -> [str]:
        """
        Return all unique nuclides for the whole database.
        Returns data as numpy array.
        Abstract method - must be extended.

        :returns: a list of strings representing the list of unique
        nuclides in the database
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    def allnuclidesoftype(self, spectype: str = "gamma") -> [str]:
        """
        Return all unique nuclides for the whole database matching a specific
        decay type, default is "gamma".
        Returns data as numpy array.
        Abstract method - must be extended.

        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: a list of strings representing the list of
        unique nuclides in the database
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    def gettypes(self, nuclide: str) -> List[str]:
        """
        Return all unique spectral types for a given radionuclide in the database.
        Abstract method - must be extended.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :returns: a list of strings representing the list of unique spcetral
        types for that nuclide
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    def hastype(self, nuclide: str, spectype: str = "gamma") -> bool:
        """
        Check if it has that particular decay type
        Abstract method - must be extended.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: boolean - true if in database, false otherwise
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    def getname(self, zai: int) -> str:
        """
        Get the name of a nuclide, given a ZAI.
        ZAI = Z (charge), A (mass number), I (isomeric state)
        Abstract method - must be extended.

        for example.
        ```
            getname(10030) = 'H3'
            getname(922351) = 'U235m'
        ```

        :param zai: the ZAI number (integer) for the nuclide
        :returns: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    def getzai(self, nuclide: str) -> int:
        """
        Get the ZAI of a nuclide, given a nuclide name.
        ZAI = Z (charge), A (mass number), I (isomeric state)
        Abstract method - must be extended.

        for example.
        ```
            getzai('H3') = 10030
            getzai('U235m') = 922351
        ```

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :returns: the ZAI number (integer) for the nuclide
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    def gethalflife(self, nuclide: str) -> float:
        """
        Get the halflife of a given nuclide in seconds
        Abstract method - must be extended.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :returns: the half life in seconds
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    @asarray
    def getenergies(self, nuclide: str, spectype: str = "gamma") -> List[float]:
        """
        Get the line energies of a given nuclide in eV.
        Default spectral type is "gamma"
        Only provides discrete lines.
        If nuclide has no spectral data then returns an empty array.
        Abstract method - must be extended.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: a numpy array of energies (eV) for the given nuclide
        and decay type.
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)

    @asarray
    def getintensities(self, nuclide: str, spectype: str = "gamma") -> List[float]:
        """
        Get the corresponding intensity values for each line energy of a
        given nuclide.
        Default spectral type is "gamma".
        Intensity values will be between 0 and 1.
        Only provides intensities (probabilities) for discrete lines.
        Also multiplies by normalisation constant.
        If nuclide has no spectral data then returns an empty array.
        Abstract method - must be extended.

        This array will be of the same size as getenergies

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: a numpy array of normalised intensities for the given
        nuclide and decay type.
        :raises AbstractClassException: raises an exception if called
        """
        raise AbstractClassException(ABSTRACT_STR_ERROR)


# a facade layer to interact with database
class DefaultDatabase(ReadOnlyDatabase):
    """
    A simple read only database for interacting with data

    This assumes the datasource is a JSON file loader, conforming
    to the correct schema

    ```
    {
        'H3': {
            'halflife': 389105000.0,
            'zai': 10030,
            'beta': {
                'lines': {
                    'energies': [18571.0],
                    'energies_unc': [6.0],
                    'intensities': [1.0],
                    'intensities_unc': [0.0],
                    'norms': [1.0],
                    'norms_unc': [0.0]
                },
                'mean_energy': 5707.4,
                'mean_energy_unc': 1.84397,
                'mean_normalisation': 1.0,
                'mean_normalisation_unc': 0.0,
                'number': 1
            },
        }
    }
    ```
    """

    IGNORE_KEYS = ["zai", "halflife"]

    def __init__(self, *args, **kwargs):
        """
        Construct the database given a datasource

        ```
        with datasource as source:
            # do something with source here
            ...
        ```

        :param datasource: context manager to load data into database
        """
        ReadOnlyDatabase.__init__(self, *args, **kwargs)

    def __contains__(self, nuclide: str) -> bool:
        """
        Check if nuclide exists in database

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :returns: boolean - true if in database, false otherwise
        """
        return nuclide in self.raw

    @property
    @sortresult
    def alltypes(self) -> List[str]:
        """
        Return all possible unique decay mode types for the whole database

        :returns: a list of strings representing the list of decay types
        in the database
        """
        spectypes = []
        for _, v in self.raw.items():
            for key in v.keys():
                if key not in self.IGNORE_KEYS and key not in spectypes:
                    spectypes.append(key)
        return spectypes

    def haslines(self, nuclide: str, spectype: str = "gamma") -> bool:
        """
        Return true if nuclide has spectral information, false otherwise

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: boolean based on if spectral data for type exists and
        has data in database
        :raises KeyError: raises an exception if nuclide and spectype not in database
        """
        return "lines" in self.raw[nuclide][spectype]

    @property
    def allnuclides(self) -> List[str]:
        """
        Return all unique nuclides for the whole database.
        Returns data as numpy array.

        :returns: a list of strings representing the list of unique nuclides
        in the database
        """
        return [k for k, _ in self.raw.items()]

    def allnuclidesoftype(self, spectype: str = "gamma") -> List[str]:
        """
        Return all unique nuclides for the whole database matching a specific
        decay type, default is "gamma".
        Returns data as numpy array.

        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: a list of strings representing the list of unique nuclides
        in the database
        :raises KeyError: raises an exception if spectype not in database
        """
        return [k for k, _ in self.raw.items() if spectype in self.raw[k].keys()]

    @sortresult
    def gettypes(self, nuclide: str) -> List[str]:
        """
        Return all unique spectral types for a given radionuclide in the database.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :returns: a list of strings representing the list of unique spcetral
        types for that nuclide
        :raises KeyError: raises an exception if nuclide not in database
        """
        return sorted(
            [k for k, _ in self.raw[nuclide].items() if k not in self.IGNORE_KEYS]
        )

    def hastype(self, nuclide: str, spectype: str = "gamma") -> bool:
        """
        Check if it has that particular decay type
        Abstract method - must be extended.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: boolean - true if in database, false otherwise
        :raises KeyError: raises an exception if nuclide not in database
        """
        return spectype in self.raw[nuclide]

    def getname(self, zai: int) -> str:
        """
        Get the name of a nuclide, given a ZAI.
        ZAI = Z (charge), A (mass number), I (isomeric state)

        for example.
        ```
            getname(10030) = 'H3'
            getname(922351) = 'U235m'
        ```

        :param zai: the ZAI number (integer) for the nuclide
        :returns: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        """
        for k, v in self.raw.items():
            if v["zai"] == zai:
                return k
        return None

    def getzai(self, nuclide: str) -> int:
        """
        Get the ZAI of a nuclide, given a nuclide name.
        ZAI = Z (charge), A (mass number), I (isomeric state)
        Abstract method - must be extended.

        for example.
        ```
            getzai('H3') = 10030
            getzai('U235m') = 922351
        ```

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :returns: the ZAI number (integer) for the nuclide
        :raises KeyError: raises an exception if nuclide key not in database
        """
        return self.raw[nuclide]["zai"]

    def gethalflife(self, nuclide: str) -> float:
        """
        Get the halflife of a given nuclide in seconds
        Abstract method - must be extended.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :returns: the half life in seconds
        :raises KeyError: raises an exception if nuclide key not in database
        """
        return self.raw[nuclide]["halflife"]

    @asarray
    def getenergies(self, nuclide: str, spectype: str = "gamma") -> List[float]:
        """
        Get the line energies of a given nuclide in eV.
        Default spectral type is "gamma"
        Only provides discrete lines.

        If nuclide has no spectral data then returns an empty array.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: a numpy array of energies (eV) for the given nuclide
        and decay type.
        :raises KeyError: raises an exception if nuclide and spectype not
        in database
        """
        if self.haslines(nuclide, spectype=spectype):
            return self.raw[nuclide][spectype]["lines"]["energies"]
        return []

    @asarray
    def getenergiesunc(self, nuclide: str, spectype: str = "gamma") -> List[float]:
        """
        Get the line energies uncertainties of a given nuclide in eV.
        Default spectral type is "gamma"
        Only provides discrete lines.

        If nuclide has no spectral data then returns an empty array.

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: a numpy array of energies (eV) for the given nuclide
        and decay type.
        :raises KeyError: raises an exception if nuclide and spectype not
        in database
        """
        if self.haslines(nuclide, spectype=spectype):
            return self.raw[nuclide][spectype]["lines"]["energies_unc"]
        return []

    @asarray
    def getintensities(self, nuclide: str, spectype: str = "gamma") -> List[float]:
        """
        Get the corresponding intensity values for each line energy of a
        given nuclide.
        Default spectral type is "gamma".
        Intensity values will be between 0 and 1.
        Only provides intensities (probabilities) for discrete lines.
        Also multiplies by normalisation constant.

        If nuclide has no spectral data then returns an empty array.

        This array will be of the same size as getenergies

        :param nuclide: the radionuclide as a string i.e 'H3' or 'U235m'.
        No spaces and case sensitive!
        :param spectype: a string representing the type of decay mode.
        Gamma is default.
        :returns: a numpy array of normalised intensities for the given
        nuclide and decay type.
        :raises KeyError: raises an exception if nuclide and spectype
        not in database
        """
        if self.haslines(nuclide, spectype=spectype):
            return [
                intensity * self.raw[nuclide][spectype]["lines"]["norms"][i]
                for i, intensity in enumerate(
                    self.raw[nuclide][spectype]["lines"]["intensities"]
                )
            ]
        return []


# some aliases for decay data
def Decay2012Database():
    """
    Alias/factory for decay 2012 data
    """
    return DefaultDatabase(
        datasource=DatabaseJSONFileLoader(datafile=__RAW_DATABASE_DECAY_2012_FILE__)
    )


def sortedlines(
    db: ReadOnlyDatabase, spectype: str = "gamma", byenergy: bool = True
) -> List[Tuple[str, float]]:
    """
    Get a sorted list (by increasing line energy) of nuclides
    with energies.

    :param db: the database holding line energies
    :param spectype: a string representing the type of decay mode.
    Gamma is default.
    :param byenergy: by default always sort by energy, otherwise sort
    alphabetically on nuclide name.
    :returns: a list of str,float pairs with the first representing the
    nuclide name, and the second representing the energy of the line

    ```
        db = ag.Decay2012Database()
        gammalines = sortedlines(db, spectype="gamma")

        # gammalines
        >>> [('U235m', 76.8), ('Es254', 1100.0), ('Ag110m', 1160.0),
            ('Tl201', 1580.0), ('Pt193m', 1642.0), ...]
    ```
    """

    # get all nuclides of type spectype
    allnuclides = db.allnuclidesoftype(spectype=spectype)

    # all lines
    alllines = []
    for nuc in allnuclides:
        lines = db.getenergies(nuc, spectype=spectype)
        nuclides = [nuc] * len(lines)
        alllines.extend(zip(nuclides, lines))

    # sort by energy
    key = 1 if byenergy else 0
    return sorted(alllines, key=lambda x: x[key])



"""
    The complete actigamma package

    ```
    import actigamma as ag

    db = ag.Decay2012Database()

    ...
    ```
"""
from .database import *
from .decorators import *
from .core import *
from .exceptions import *
from .identifier import *
from .inventory import *
from .util import *

# version in two places - here and .VERSION file
__version__ = "0.1.5"



import os
from setuptools import setup


VERSION = ""
with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), '.VERSION'), 'rt') as vfile:
      VERSION = vfile.readlines()[0].strip("\n").strip()

setup(name='actigamma',
      version=VERSION,
      description='The package for producing gamma spec from nuclide activities',
      long_description_content_type='text/x-rst',
      long_description='The package for producing gamma spec from nuclide activities',
      url='https://github.com/fispact/actigamma',
      author='UKAEA',
      author_email='thomas.stainer@ukaea.uk',
      license='Apache License 2.0',
      packages=[
            'actigamma',
      ],
      install_requires=[],
      python_requires='>=3',
      scripts=[],
      setup_requires=[
            'pytest-runner',
      ],
      test_suite='tests.testsuite',
      tests_require=[
            'pytest',
            'pytest-cov>=2.3.1',
            'mock',
            'numpy',
      ],
      include_package_data=True,
      zip_safe=False,
      )
